package cn.com.qmth.examcloud.core.basic.api.controller;

import cn.com.qmth.examcloud.api.commons.enums.CourseLevel;
import cn.com.qmth.examcloud.api.commons.enums.DataRuleType;
import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.api.commons.security.bean.UserDataRule;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.helpers.poi.ExcelWriter;
import cn.com.qmth.examcloud.commons.util.PathUtil;
import cn.com.qmth.examcloud.core.basic.api.CourseCloudService;
import cn.com.qmth.examcloud.core.basic.api.controller.bean.CourseDomain;
import cn.com.qmth.examcloud.core.basic.dao.CourseRepo;
import cn.com.qmth.examcloud.core.basic.dao.entity.CourseEntity;
import cn.com.qmth.examcloud.core.basic.dao.entity.CourseSpeciatlyRelationEntity;
import cn.com.qmth.examcloud.core.basic.service.CourseService;
import cn.com.qmth.examcloud.core.basic.service.UserDataRuleService;
import cn.com.qmth.examcloud.core.basic.service.bean.CourseInfo;
import cn.com.qmth.examcloud.task.api.DataSyncCloudService;
import cn.com.qmth.examcloud.task.api.request.SyncCourseReq;
import cn.com.qmth.examcloud.web.config.SystemProperties;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import cn.com.qmth.examcloud.web.security.DataRule;
import cn.com.qmth.examcloud.web.support.ControllerSupport;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.fileupload.disk.DiskFileItem;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import org.springframework.data.domain.Sort.Direction;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.commons.CommonsMultipartFile;

import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;
import javax.persistence.criteria.Subquery;
import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * 课程服务API Created by songyue on 17/1/14.
 */
@RestController
@Api(tags = "课程相关接口")
@RequestMapping("${$rmp.ctr.basic}/course")
public class CourseController extends ControllerSupport {

    @Autowired
    CourseRepo courseRepo;

    @Autowired
    SystemProperties systemConfig;

    @Autowired
    CourseService courseService;

    @Autowired
    CourseCloudService courseCloudService;

    @Autowired
    DataSyncCloudService dataSyncCloudService;

    @Autowired
    UserDataRuleService userDataRuleService;

    private static final String[] EXCEL_HEADER = new String[]{"课程名称", "课程代码", "层次(ZSB,GQZ,ALL)"};

    /**
     * 方法注释
     *
     * @param curPage
     * @param pageSize
     * @param name
     * @param code
     * @param level
     * @param enable
     * @param specialtyId
     * @return
     * @author WANGWEI
     */
    @DataRule(type = DataRuleType.COURSE)
    @ApiOperation(value = "分页查询课程")
    @GetMapping("coursePage/{curPage}/{pageSize}")
    public Page<CourseEntity> getCoursePage(@PathVariable Integer curPage,
                                            @PathVariable Integer pageSize,
                                            @RequestParam(required = false) String name,
                                            @RequestParam(required = false) String code,
                                            @RequestParam(required = false) String level,
                                            @RequestParam(required = false) Boolean enable,
                                            @RequestParam(required = false) Long specialtyId) {
        User accessUser = getAccessUser();

        PageRequest pageable = PageRequest.of(curPage, pageSize, Sort.by(Direction.DESC, "updateTime", "id"));

        UserDataRule userDataRule = super.getUserDataRule(DataRuleType.COURSE);
        if (userDataRule.assertEmptyQueryResult()) {
            return new PageImpl<>(new ArrayList<>(), pageable, 0L);
        }

        Specification<CourseEntity> specification = (root, query, cb) -> {
            List<Predicate> predicates = new ArrayList<>();

            if (userDataRule.assertNeedQueryRefIds()) {
                predicates.add(root.get("id").in(userDataRule.getRefIds()));
            }

            predicates.add(cb.equal(root.get("rootOrgId"), accessUser.getRootOrgId()));

            if (StringUtils.isNotBlank(name)) {
                predicates.add(cb.like(root.get("name"), toSqlSearchPattern(name)));
            }
            if (StringUtils.isNotBlank(code)) {
                predicates.add(cb.like(root.get("code"), toSqlSearchPattern(code)));
            }
            if (StringUtils.isNotBlank(level)) {
                predicates.add(cb.equal(root.get("level"), CourseLevel.valueOf(level)));
            }
            if (null != enable) {
                predicates.add(cb.equal(root.get("enable"), enable));
            }

            if (null != specialtyId) {
                Subquery<CourseSpeciatlyRelationEntity> subquery = query
                        .subquery(CourseSpeciatlyRelationEntity.class);
                Root<CourseSpeciatlyRelationEntity> subRoot = subquery
                        .from(CourseSpeciatlyRelationEntity.class);
                subquery.select(subRoot.get("courseId"));
                Predicate p1 = cb.equal(subRoot.get("specialtyId"), specialtyId);
                Predicate p2 = cb.equal(subRoot.get("courseId"), root.get("id"));
                subquery.where(cb.and(p1, p2));
                predicates.add(cb.exists(subquery));
            }

            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };

        Page<CourseEntity> page = courseRepo.findAll(specification, pageable);
        return page;
    }

    /**
     * 方法注释
     *
     * @param name
     * @param level
     * @param enable
     * @param specialtyId
     * @return
     * @author WANGWEI
     */
    @DataRule(type = DataRuleType.COURSE)
    @ApiOperation(value = "查询课程")
    @GetMapping("query")
    public List<CourseEntity> query(@RequestParam(required = false) Long rootOrgId,
                                    @RequestParam(required = false) String name,
                                    @RequestParam(required = false) String level,
                                    @RequestParam(required = false) Boolean enable,
                                    @RequestParam(required = false) Long specialtyId) {
        User accessUser = getAccessUser();
        if (null == rootOrgId) {
            rootOrgId = accessUser.getRootOrgId();
        }
        final Long finalRootOrgId = rootOrgId;

        UserDataRule userDataRule = super.getUserDataRule(DataRuleType.COURSE);
        if (userDataRule.assertEmptyQueryResult()) {
            return new ArrayList<>();
        }

        Specification<CourseEntity> specification = (root, query, cb) -> {
            List<Predicate> predicates = new ArrayList<>();

            if (userDataRule.assertNeedQueryRefIds()) {
                predicates.add(root.get("id").in(userDataRule.getRefIds()));
            }

            predicates.add(cb.equal(root.get("rootOrgId"), finalRootOrgId));
            Predicate pr1 = cb.like(root.get("name"), toSqlSearchPattern(name));
            Predicate pr2 = cb.like(root.get("code"), toSqlSearchPattern(name));
            predicates.add(cb.or(pr1, pr2));

            if (StringUtils.isNotBlank(level)) {
                predicates.add(cb.equal(root.get("level"), toSqlSearchPattern(level)));
            }

            if (null != enable) {
                predicates.add(cb.equal(root.get("enable"), enable));
            }

            if (null != specialtyId) {
                Subquery<CourseSpeciatlyRelationEntity> subquery = query.subquery(CourseSpeciatlyRelationEntity.class);
                Root<CourseSpeciatlyRelationEntity> subRoot = subquery.from(CourseSpeciatlyRelationEntity.class);
                subquery.select(subRoot.get("courseId"));
                Predicate p1 = cb.equal(subRoot.get("specialtyId"), specialtyId);
                Predicate p2 = cb.equal(subRoot.get("courseId"), root.get("id"));
                subquery.where(cb.and(p1, p2));
                predicates.add(cb.exists(subquery));
            }

            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };

        PageRequest pageRequest = PageRequest.of(0, 50, Sort.by(Direction.DESC, "updateTime"));
        Page<CourseEntity> page = courseRepo.findAll(specification, pageRequest);

        Iterator<CourseEntity> iterator = page.iterator();
        List<CourseEntity> list = Lists.newArrayList();

        while (iterator.hasNext()) {
            CourseEntity next = iterator.next();

            CourseEntity bean = new CourseEntity();
            bean.setCode(next.getCode());
            bean.setEnable(next.getEnable());
            bean.setId(next.getId());
            bean.setLevel(next.getLevel());
            bean.setName(next.getName());
            bean.setRootOrgId(next.getRootOrgId());

            list.add(bean);
        }

        return list;
    }

    /**
     * 方法注释
     *
     * @param id
     * @return
     * @author WANGWEI
     */
    @ApiOperation(value = "按ID查询课程", notes = "ID查询")
    @GetMapping("{id}")
    public CourseEntity getCourseById(@PathVariable Long id) {
        CourseEntity course = GlobalHelper.getEntity(courseRepo, id, CourseEntity.class);
        if (null == course) {
            throw new StatusException("540001", "课程不存在");
        }

        Long rootOrgId = course.getRootOrgId();
        validateRootOrgIsolation(rootOrgId);
        return course;
    }

    /**
     * 方法注释
     *
     * @param rootOrgId
     * @param code
     * @return
     * @author WANGWEI
     */
    @ApiOperation(value = "按code查询课程", notes = "code查询")
    @GetMapping("byCode")
    public CourseEntity getCourseByCode(@RequestParam Long rootOrgId, @RequestParam String code) {
        validateRootOrgIsolation(rootOrgId);

        CourseEntity course = courseRepo.findByRootOrgIdAndCode(rootOrgId, code);
        if (null == course) {
            throw new StatusException("540001", "课程不存在");
        }

        return course;
    }

    /**
     * 修正
     *
     * @param domain
     * @return
     * @author WANGWEI
     */
    @ApiOperation(value = "新增课程", notes = "新增")
    @PostMapping
    @Transactional
    public Long addCourse(@RequestBody CourseDomain domain) {
        trim(domain, true);

        User accessUser = getAccessUser();
        Long rootOrgId = accessUser.getRootOrgId();

        String code = domain.getCode();
        String name = domain.getName();
        if (StringUtils.isBlank(code)) {
            throw new StatusException("620001", "课程编码不能为空");
        }
        if (StringUtils.isBlank(name)) {
            throw new StatusException("620002", "课程名称不能为空");
        }
        if (name.length() > 50) {
            throw new StatusException("620003", "课程名称不能超过50个字符");
        }
        CourseEntity course = courseRepo.findByRootOrgIdAndCode(rootOrgId, code);
        if (null != course) {
            throw new StatusException("620002", "课程编码已被占用");
        }

        CourseInfo info = new CourseInfo();
        info.setRootOrgId(rootOrgId);
        info.setCode(domain.getCode());
        info.setEnable(domain.getEnable());
        info.setId(domain.getId());
        info.setLevel(domain.getLevel());
        info.setName(domain.getName());

        CourseEntity saved = courseService.saveCourse(info);

        // 用户与课程的数据权限
        userDataRuleService.addUserDataRule(accessUser.getUserId(), DataRuleType.COURSE, saved.getId());

        return saved.getId();
    }

    /**
     * 修正
     *
     * @param domain
     * @return
     * @author WANGWEI
     */
    @ApiOperation(value = "更新课程", notes = "更新")
    @PutMapping
    @Transactional
    public Long updateCourse(@RequestBody CourseDomain domain) {
        trim(domain, true);

        User accessUser = getAccessUser();
        Long rootOrgId = accessUser.getRootOrgId();

        String code = domain.getCode();
        String name = domain.getName();
        if (StringUtils.isBlank(code)) {
            throw new StatusException("620001", "课程编码不能为空");
        }
        if (StringUtils.isBlank(name)) {
            throw new StatusException("620002", "课程名称不能为空");
        }
        if (name.length() > 50) {
            throw new StatusException("620003", "课程名称不能超过50个字符");
        }

        CourseInfo info = new CourseInfo();
        info.setRootOrgId(rootOrgId);
        info.setCode(domain.getCode());
        info.setEnable(domain.getEnable());
        info.setId(domain.getId());
        info.setLevel(domain.getLevel());
        info.setName(domain.getName());

        CourseEntity saved = courseService.saveCourse(info);
        return saved.getId();
    }

    @ApiOperation(value = "删除课程", notes = "删除")
    @DeleteMapping("{ids}")
    @Transactional
    public void deleteCourse(@PathVariable String ids) {
        List<Long> courseIds = Stream.of(ids.split(",")).map(s -> Long.parseLong(s.trim()))
                .collect(Collectors.toList());
        for (Long courseId : courseIds) {
            CourseEntity one = GlobalHelper.getEntity(courseRepo, courseId, CourseEntity.class);
            if (null == one) {
                continue;
            }
            validateRootOrgIsolation(one.getRootOrgId());
            courseRepo.delete(one);
        }
    }

    @ApiOperation(value = "下载导入模板", notes = "下载导入模板")
    @GetMapping("importTemplate")
    public void getImportTemplate(HttpServletResponse response) {
        String resoucePath = PathUtil.getResoucePath("templates/courseImportTemplate.xlsx");
        exportFile("课程导入模板.xlsx", new File(resoucePath));
    }

    @ApiOperation(value = "导入课程", notes = "导入")
    @PostMapping("import")
    @Transactional
    public Map<String, Object> importCourse(@RequestParam CommonsMultipartFile file) {
        DiskFileItem item = (DiskFileItem) file.getFileItem();
        File storeLocation = item.getStoreLocation();
        List<Map<String, Object>> failRecords = courseService.importCourse(getRootOrgId(),
                storeLocation);
        Map<String, Object> map = Maps.newHashMap();
        map.put("hasError", CollectionUtils.isNotEmpty(failRecords));
        map.put("failRecords", failRecords);
        return map;
    }

    @DataRule(type = DataRuleType.COURSE)
    @ApiOperation(value = "导出课程", notes = "导出")
    @GetMapping("export")
    public void export(@RequestParam(required = false) String name,
                       @RequestParam(required = false) String code,
                       @RequestParam(required = false) String level,
                       @RequestParam(required = false) Boolean enable,
                       @RequestParam(required = false) Long specialtyId) {

        User accessUser = getAccessUser();

        UserDataRule userDataRule = super.getUserDataRule(DataRuleType.COURSE);
        if (userDataRule.assertEmptyQueryResult()) {
            this.doExport(new ArrayList<>());
            return;
        }

        Specification<CourseEntity> specification = (root, query, cb) -> {
            List<Predicate> predicates = new ArrayList<>();

            if (userDataRule.assertNeedQueryRefIds()) {
                predicates.add(root.get("id").in(userDataRule.getRefIds()));
            }

            predicates.add(cb.equal(root.get("rootOrgId"), accessUser.getRootOrgId()));

            if (StringUtils.isNotBlank(name)) {
                predicates.add(cb.like(root.get("name"), toSqlSearchPattern(name)));
            }
            if (StringUtils.isNotBlank(code)) {
                predicates.add(cb.like(root.get("code"), toSqlSearchPattern(code)));
            }
            if (StringUtils.isNotBlank(level)) {
                predicates.add(cb.equal(root.get("level"), CourseLevel.valueOf(level)));
            }
            if (null != enable) {
                predicates.add(cb.equal(root.get("enable"), enable));
            }

            if (null != specialtyId) {
                Subquery<CourseSpeciatlyRelationEntity> subquery = query
                        .subquery(CourseSpeciatlyRelationEntity.class);
                Root<CourseSpeciatlyRelationEntity> subRoot = subquery
                        .from(CourseSpeciatlyRelationEntity.class);
                subquery.select(subRoot.get("courseId"));
                Predicate p1 = cb.equal(subRoot.get("specialtyId"), specialtyId);
                Predicate p2 = cb.equal(subRoot.get("courseId"), root.get("id"));
                subquery.where(cb.and(p1, p2));
                predicates.add(cb.exists(subquery));
            }

            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };

        long count = courseRepo.count(specification);
        if (100000 < count) {
            throw new StatusException("620200", "数据量过大,无法导出");
        }

        List<CourseEntity> list = courseRepo.findAll(specification);
        this.doExport(list);
    }

    private void doExport(List<CourseEntity> list) {
        List<Object[]> datas = Lists.newArrayList();

        for (CourseEntity cur : list) {
            datas.add(new Object[]{cur.getName(), cur.getCode(), cur.getLevel().getName()});
        }

        String filePath = systemConfig.getTempDataDir() + File.separator
                + System.currentTimeMillis() + ".xlsx";
        File file = new File(filePath);

        ExcelWriter.write(EXCEL_HEADER, new Class[]{String.class, String.class, String.class},
                datas, new File(filePath));

        exportFile("课程列表-" + getRootOrgId() + ".xlsx", file);

        FileUtils.deleteQuietly(file);
    }

    /**
     * 方法注释
     *
     * @param ids
     * @return
     * @author WANGWEI
     */
    @ApiOperation(value = "禁用课程", notes = "禁用")
    @PutMapping("disable/{ids}")
    @Transactional
    public List<Long> disableCourse(@PathVariable String ids) {
        List<Long> courseIds = Stream.of(ids.split(",")).map(s -> Long.parseLong(s.trim()))
                .collect(Collectors.toList());

        List<CourseEntity> savedList = Lists.newArrayList();
        for (Long courseId : courseIds) {
            CourseEntity course = GlobalHelper.getEntity(courseRepo, courseId, CourseEntity.class);
            course.setEnable(false);
            CourseEntity saved = courseRepo.saveAndFlush(course);
            savedList.add(saved);
        }

        for (CourseEntity ce : savedList) {
            SyncCourseReq req = new SyncCourseReq();
            req.setCode(ce.getCode());
            req.setEnable(ce.getEnable());
            req.setId(ce.getId());
            req.setLevel(ce.getLevel().name());
            req.setName(ce.getName());
            req.setRootOrgId(ce.getRootOrgId());
            req.setSyncType("update");
            dataSyncCloudService.syncCourse(req);
        }

        return courseIds;
    }

    /**
     * 方法注释
     *
     * @param ids
     * @return
     * @author WANGWEI
     */
    @ApiOperation(value = "启用课程", notes = "启用")
    @PutMapping("enable/{ids}")
    @Transactional
    public List<Long> enableCourse(@PathVariable String ids) {
        List<Long> courseIds = Stream.of(ids.split(",")).map(s -> Long.parseLong(s.trim()))
                .collect(Collectors.toList());

        List<CourseEntity> savedList = Lists.newArrayList();
        for (Long courseId : courseIds) {
            CourseEntity course = GlobalHelper.getEntity(courseRepo, courseId, CourseEntity.class);
            course.setEnable(true);
            CourseEntity saved = courseRepo.saveAndFlush(course);
            savedList.add(saved);
        }

        for (CourseEntity ce : savedList) {
            SyncCourseReq req = new SyncCourseReq();
            req.setCode(ce.getCode());
            req.setEnable(ce.getEnable());
            req.setId(ce.getId());
            req.setLevel(ce.getLevel().name());
            req.setName(ce.getName());
            req.setRootOrgId(ce.getRootOrgId());
            req.setSyncType("update");
            dataSyncCloudService.syncCourse(req);
        }

        return courseIds;
    }

}
