package cn.com.qmth.examcloud.core.basic.service.impl;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

import cn.com.qmth.examcloud.api.commons.enums.CourseLevel;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.helpers.poi.ExcelReader;
import cn.com.qmth.examcloud.commons.util.PathUtil;
import cn.com.qmth.examcloud.core.basic.dao.CourseRepo;
import cn.com.qmth.examcloud.core.basic.dao.OrgRepo;
import cn.com.qmth.examcloud.core.basic.dao.entity.CourseEntity;
import cn.com.qmth.examcloud.core.basic.dao.entity.OrgEntity;
import cn.com.qmth.examcloud.core.basic.service.CourseService;
import cn.com.qmth.examcloud.core.basic.service.bean.CourseInfo;
import cn.com.qmth.examcloud.core.basic.service.cache.CourseCache;
import cn.com.qmth.examcloud.task.api.DataSyncCloudService;
import cn.com.qmth.examcloud.task.api.request.SyncCourseReq;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;

@Service
public class CourseServiceImpl implements CourseService {

	@Autowired
	CourseRepo courseRepo;

	@Autowired
	OrgRepo orgRepo;

	@Autowired
	DataSyncCloudService dataSyncCloudService;

	@Autowired
	CourseCache courseCache;

	private static final String[] EXCEL_HEADER = new String[]{"课程名称", "课程代码", "层次(ZSB,GQZ,ALL)"};

	/*
	 * 实现
	 *
	 * @author WANGWEI
	 * 
	 * @see
	 * cn.com.qmth.examcloud.core.basic.service.CourseService#saveCourse(cn.com.
	 * qmth.examcloud.core.basic.service.bean.CourseInfo)
	 */
	@Override
	public CourseEntity saveCourse(CourseInfo courseInfo) {
		Long rootOrgId = courseInfo.getRootOrgId();
		Long courseId = courseInfo.getId();
		String courseCode = courseInfo.getCode();
		CourseLevel level = courseInfo.getLevel();
		Boolean enable = courseInfo.getEnable();
		String name = courseInfo.getName();

		if (null == rootOrgId) {
			throw new StatusException("160000", "rootOrgId is null");
		}
		OrgEntity rootOrg = GlobalHelper.getEntity(orgRepo, rootOrgId, OrgEntity.class);
		if (null == rootOrg) {
			throw new StatusException("160001", "机构不存在");
		}
		if (null != rootOrg.getParentId()) {
			throw new StatusException("160002", "机构错误");
		}

		CourseEntity course = null;

		if (null != courseId) {
			course = GlobalHelper.getEntity(courseRepo, courseId, CourseEntity.class);

			if (null == course) {
				throw new StatusException("160006", "课程ID错误");
			}

			if (!rootOrgId.equals(course.getRootOrgId())) {
				throw new StatusException("160006", "课程机构错误");
			}

		} else if (StringUtils.isNotBlank(courseCode)) {

			course = courseRepo.findByRootOrgIdAndCode(rootOrgId, courseCode);

			if (null != course) {
				if (null == course.getEnable()) {
					course.setEnable(true);
				}
				if (null == course.getLevel()) {
					course.setLevel(CourseLevel.ALL);
				}
			} else {
				if (StringUtils.isBlank(name)) {
					throw new StatusException("160003", "courseName is null");
				}

				course = new CourseEntity();
				course.setRootOrgId(rootOrgId);
				course.setCode(courseCode);
				course.setEnable(true);
				course.setLevel(CourseLevel.ALL);
			}

		} else {
			throw new StatusException("160004", "courseId,courseCode can not be all null");
		}

		if (null != enable) {
			course.setEnable(enable);
		}

		if (StringUtils.isNotBlank(name)) {
			course.setName(name);
		}

		if (null != level) {
			course.setLevel(level);
		}

		long updateTime = 0L;
		if (null != course.getUpdateTime()) {
			updateTime = course.getUpdateTime().getTime();
		}

		CourseEntity saved = courseRepo.saveAndFlush(course);

		// 同步操作
		if (updateTime != saved.getUpdateTime().getTime()) {
			SyncCourseReq req = new SyncCourseReq();
			req.setCode(saved.getCode());
			req.setEnable(saved.getEnable());
			req.setId(saved.getId());
			req.setLevel(saved.getLevel().name());
			req.setName(saved.getName());
			req.setRootOrgId(saved.getRootOrgId());
			req.setSyncType("update");
			dataSyncCloudService.syncCourse(req);
		}

		courseCache.remove(saved.getId());

		return saved;
	}

	@Override
	public List<Map<String, Object>> importCourse(Long rootOrgId, File file) {

		List<String[]> lineList = null;
		try {
			lineList = ExcelReader.readSheetBySax(PathUtil.getCanonicalPath(file), 1, 3);
		} catch (Exception e) {
			throw new StatusException("100110", "Excel 解析失败");
		}

		if (CollectionUtils.isEmpty(lineList)) {
			throw new StatusException("100111", "Excel无内容");
		}

		if (10001 < lineList.size()) {
			throw new StatusException("100112", "数据行数不能超过10000");
		}

		List<Map<String, Object>> failRecords = Collections
				.synchronizedList(new ArrayList<Map<String, Object>>());

		List<CourseInfo> courseList = Lists.newArrayList();

		for (int i = 0; i < lineList.size(); i++) {
			String[] line = lineList.get(i);
			if (0 == i) {
				if (headerError(line)) {
					throw new StatusException("100111", "Excel表头错误");
				}
				continue;
			}

			boolean hasError = false;
			StringBuilder msg = new StringBuilder();

			CourseInfo couse = new CourseInfo();
			couse.setRootOrgId(rootOrgId);
			couse.setEnable(true);

			String name = trimAndNullIfBlank(line[0]);
			if (StringUtils.isBlank(name)) {
				msg.append("  课程名称不能为空");
				hasError = true;
			} else if (name.length() > 30) {
				msg.append("  课程名称不能超过30个字符");
				hasError = true;
			}
			couse.setName(name);

			String code = trimAndNullIfBlank(line[1]);
			if (StringUtils.isBlank(code)) {
				msg.append("  课程代码不能为空");
				hasError = true;
			} else if (code.length() > 30) {
				msg.append("  课程代码不能超过30个字符");
				hasError = true;
			}

			couse.setCode(code);

			String level = trimAndNullIfBlank(line[2]);
			if (StringUtils.isBlank(level)) {
				msg.append("  课程层次不能为空");
				hasError = true;
			} else {
				try {
					CourseLevel cl = CourseLevel.getCourseLevel(level);
					couse.setLevel(cl);
				} catch (Exception e) {
					msg.append("  课程层次错误");
					hasError = true;
				}
			}

			if (hasError) {
				failRecords.add(newError(i + 1, msg.toString()));
			} else {
				courseList.add(couse);
			}

		}

		if (CollectionUtils.isNotEmpty(failRecords)) {
			return failRecords;
		}

		for (CourseInfo cur : courseList) {
			saveCourse(cur);
		}

		return failRecords;
	}

	private Map<String, Object> newError(int lineNum, String msg) {
		Map<String, Object> map = Maps.newHashMap();
		map.put("lineNum", lineNum);
		map.put("msg", msg);
		return map;
	}

	private String trimAndNullIfBlank(String s) {
		if (StringUtils.isBlank(s)) {
			return null;
		}
		return s.trim();
	}

	/**
	 * 方法注释
	 *
	 * @author WANGWEI
	 * @param header
	 */
	private boolean headerError(String[] header) {
		for (int i = 0; i < EXCEL_HEADER.length; i++) {
			if (null == header[i]) {
				return true;
			}
			if (!EXCEL_HEADER[i].equals(header[i].trim())) {
				return true;
			}
		}
		return false;
	}

}
