/*
 * Copyright (c) 2020 QMTH All Rights Reserved.
 * Created by Deason on 2020-08-18 15:31:08
 */

package cn.com.qmth.examcloud.core.basic.service.impl;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.commons.util.JsonUtil;
import cn.com.qmth.examcloud.core.basic.dao.LoginRuleRepo;
import cn.com.qmth.examcloud.core.basic.dao.entity.LoginRuleEntity;
import cn.com.qmth.examcloud.core.basic.dao.enums.LoginRuleType;
import cn.com.qmth.examcloud.core.basic.service.LoginRuleService;
import cn.com.qmth.examcloud.core.basic.service.OrgService;
import cn.com.qmth.examcloud.core.basic.service.bean.LoginRuleForm;
import cn.com.qmth.examcloud.core.basic.service.bean.LoginRuleInfo;
import cn.com.qmth.examcloud.core.basic.service.bean.LoginRuleQuery;
import cn.com.qmth.examcloud.support.filestorage.FileStorageUtil;
import cn.com.qmth.examcloud.web.filestorage.FileStoragePathEnvInfo;
import cn.com.qmth.examcloud.web.filestorage.YunPathInfo;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.MapUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.*;
import org.springframework.data.jpa.domain.Specification;
import org.springframework.stereotype.Service;

import javax.persistence.criteria.Predicate;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class LoginRuleServiceImpl implements LoginRuleService {

    private static final Logger LOG = LoggerFactory.getLogger(LoginRuleServiceImpl.class);

    /**
     * “登录规则”的缓存KEY前缀
     */
    private static final String LOGIN_RULE_CACHE_KEY_PREFIX = "$LOGIN_RULE:";

    /**
     * ../resources/aliyun.xml siteId
     */
    private static final String LOGIN_RULE_SITE_ID = "orgProperties";

    private static final long TOP_ID = -1L;

    @Autowired
    private LoginRuleRepo loginRuleRepo;

    @Autowired
    private OrgService orgService;

    @Autowired
    private RedisClient redisClient;

    @Override
    public List<LoginRuleInfo> getLoginRuleTopList() {
        LoginRuleQuery req = new LoginRuleQuery();
        req.setPageNo(1);
        req.setPageSize(100);
        req.setRootOrgId(TOP_ID);

        return this.getLoginRuleList(req).getContent();
    }

    @Override
    public Page<LoginRuleInfo> getLoginRuleList(LoginRuleQuery req) {
        if (req.getPageNo() == null || req.getPageNo() < 1) {
            req.setPageNo(1);
        }

        if (req.getPageSize() == null || req.getPageSize() < 1) {
            req.setPageSize(10);
        }

        Specification<LoginRuleEntity> spec = (root, query, cb) -> {
            List<Predicate> predicates = new ArrayList<>();

            if (req.getRootOrgId() != null) {
                predicates.add(cb.equal(root.get("rootOrgId"), req.getRootOrgId()));
            } else {
                predicates.add(cb.notEqual(root.get("rootOrgId"), TOP_ID));
            }

            if (req.getType() != null) {
                predicates.add(cb.equal(root.get("type"), req.getType()));
            }

            return cb.and(predicates.toArray(new Predicate[predicates.size()]));
        };

        Sort sort = Sort.by(Sort.Order.desc("id"));
        Pageable pageable = PageRequest.of(req.getPageNo() - 1, req.getPageSize(), sort);
        Page<LoginRuleEntity> page = loginRuleRepo.findAll(spec, pageable);

        if (!page.hasContent()) {
            return new PageImpl<>(new ArrayList<>(), pageable, 0);
        }

        Set<Long> rootOrgIds = page.getContent().stream().filter(e -> e.getRootOrgId() >= 0L).map(e -> e.getRootOrgId()).collect(Collectors.toSet());
        Map<Long, String> rootOrgNames = orgService.findRootOrgNamesByIds(rootOrgIds);

        List<LoginRuleInfo> list = page.getContent().stream().map(entity -> {
            LoginRuleInfo info = new LoginRuleInfo();
            info.setId(entity.getId());
            info.setRootOrgId(entity.getRootOrgId());
            info.setRootOrgName(rootOrgNames.get(entity.getRootOrgId()));
            info.setType(entity.getType());
            info.setAllow(entity.getAllow());
            info.setCreationTime(entity.getCreationTime());
            info.setUpdateTime(entity.getUpdateTime());
            return info;
        }).collect(Collectors.toList());

        return new PageImpl<>(list, pageable, page.getTotalElements());
    }

    @Override
    public void addOrUpdateLoginRule(LoginRuleForm req) {
        if (req.getRootOrgId() == null) {
            throw new StatusException("400400", "顶层机构ID不能为空");
        }

        if (req.getType() == null) {
            throw new StatusException("400400", "规则类型不能为空");
        }

        if (req.getAllow() == null) {
            throw new StatusException("400400", "是否允许不能为空");
        }

        List<LoginRuleEntity> rules = loginRuleRepo.findByRootOrgIdAndType(req.getRootOrgId(), req.getType());

        LoginRuleEntity rule;
        if (CollectionUtils.isNotEmpty(rules)) {
            // 默认只存在一个值
            rule = rules.get(0);
        } else {
            rule = new LoginRuleEntity();
        }

        rule.setRootOrgId(req.getRootOrgId());
        rule.setType(req.getType());
        rule.setAllow(req.getAllow());
        loginRuleRepo.save(rule);

        this.refreshLoginRule(rule.getType());
    }

    @Override
    public void deleteLoginRuleById(Long id) {
        Optional<LoginRuleEntity> optional = loginRuleRepo.findById(id);
        if (optional.isPresent()) {
            loginRuleRepo.delete(optional.get());

            this.refreshLoginRule(optional.get().getType());
        }
    }

    @Override
    public void refreshAllLoginRule() {
        // 按规则类型分别缓存
        for (LoginRuleType type : LoginRuleType.values()) {
            this.refreshLoginRule(type);
        }
    }

    private void refreshLoginRule(LoginRuleType type) {
        List<LoginRuleEntity> entities = loginRuleRepo.findByType(type);
        Map<String, Boolean> values = new HashMap<>();
        if (CollectionUtils.isNotEmpty(entities)) {
            for (LoginRuleEntity entity : entities) {
                values.put(String.valueOf(entity.getRootOrgId()), entity.getAllow());
            }
        }

        String cacheKey = LOGIN_RULE_CACHE_KEY_PREFIX + type.name();
        redisClient.set(cacheKey, values);

        if (LoginRuleType.GEETEST_LOGIN == type) {
            String json = JsonUtil.toJson(values);
            this.uploadJsonFile(json);
        }

        LOG.info("refreshLoginRule " + type.name());
    }

    private void uploadJsonFile(String json) {
        try {
            // 路径：../org_properties/geetestConfig.json
            FileStoragePathEnvInfo env = new FileStoragePathEnvInfo();
            env.setExt1("geetestConfig");
            YunPathInfo oss = FileStorageUtil.saveFile(LOGIN_RULE_SITE_ID, env, json.getBytes(), true);
            LOG.debug(oss.getRelativePath());
        } catch (Exception e) {
            LOG.error(e.getMessage(), e);
        }
    }

    /**
     * 检查当前机构的登录规则配置
     * 1、未配置时，默认true
     * 2、优先按“当前机构设置值”，其次按“全局设置值”
     */
    @Override
    public boolean isLoginRuleAllow(Long rootOrgId, LoginRuleType type) {
        if (rootOrgId == null) {
            throw new StatusException("400400", "顶层机构ID不能为空");
        }

        if (type == null) {
            throw new StatusException("400400", "规则类型不能为空");
        }

        String cacheKey = LOGIN_RULE_CACHE_KEY_PREFIX + type.name();
        Map<String, Boolean> values = redisClient.get(cacheKey, HashMap.class);
        if (MapUtils.isEmpty(values)) {
            LOG.warn(String.format("LoginRule not exist, type = %s, rootOrgId = %s", rootOrgId, type.name()));
            return true;
        }

        // 全局设置值
        Boolean globalAllow = values.get(String.valueOf(TOP_ID));
        // 当前机构设置值
        Boolean allow = values.get(String.valueOf(rootOrgId));

        if (globalAllow == null || globalAllow) {
            if (allow == null || allow) {
                LOG.info(String.format("LoginRule allow, rootOrgId = %s, type = %s", rootOrgId, type.name()));
                return true;
            }
        } else {
            if (allow != null && allow) {
                LOG.info(String.format("LoginRule allow, rootOrgId = %s, type = %s", rootOrgId, type.name()));
                return true;
            }
        }

        LOG.info(String.format("LoginRule unAllow, rootOrgId = %s, type = %s", rootOrgId, type.name()));
        return false;
    }

}
